@@ -7,6 +7,7 @@
 #define _EVENT_H
 
 #include <stdint.h>
+#include "asprof.h"
 #include "os.h"
 
 
@@ -15,14 +16,15 @@ enum EventType {
     PERF_SAMPLE,
     EXECUTION_SAMPLE,
     WALL_CLOCK_SAMPLE,
-    INSTRUMENTED_METHOD,
     MALLOC_SAMPLE,
+    INSTRUMENTED_METHOD,
     ALLOC_SAMPLE,
     ALLOC_OUTSIDE_TLAB,
     LIVE_OBJECT,
     LOCK_SAMPLE,
     PARK_SAMPLE,
     PROFILING_WINDOW,
+    USER_EVENT,
 };
 
 class Event {
@@ -83,4 +85,12 @@ class MallocEvent : public Event {
     u64 _size;
 };
 
+class UserEvent : public Event {
+  public:
+    u64 _start_time;
+    asprof_jfr_event_key _type;
+    const uint8_t* _data;
+    size_t _len;
+};
+
 #endif // _EVENT_H
