@@ -17,6 +17,8 @@
 
 #define RESTARTABLE(call)  ({ ssize_t ret; while ((ret = call) < 0 && errno == EINTR); ret; })
 
+#define MAX_PROBE_LEN  256
+
 
 // base header for all requests
 enum request_type {
@@ -34,6 +36,7 @@ struct perf_fd_request {
     int tid;
     int target_cpu;
     struct perf_event_attr attr;
+    char probe_name[MAX_PROBE_LEN];
 };
 
 struct fd_response {
