@@ -3,6 +3,7 @@
  * SPDX-License-Identifier: Apache-2.0
  */
 
+#include <assert.h>
 #include <map>
 #include <string>
 #include <arpa/inet.h>
@@ -26,6 +27,7 @@
 #include "threadFilter.h"
 #include "threadLocalData.h"
 #include "tsc.h"
+#include "userEvents.h"
 #include "vmStructs.h"
 
 
@@ -57,7 +59,7 @@ static jmethodID _start_method;
 static jmethodID _stop_method;
 static jmethodID _box_method;
 
-static const char* const SETTING_CSTACK[] = {NULL, "no", "fp", "dwarf", "lbr", "vm"};
+static const char* const SETTING_CSTACK[] = {NULL, "no", "fp", "dwarf", "lbr", "vm", "vmx"};
 
 
 struct CpuTime {
@@ -172,31 +174,24 @@ class Lookup {
     }
 
     bool fillJavaMethodInfo(MethodInfo* mi, jmethodID method, bool first_time) {
-        if (VMStructs::hasMethodStructs()) {
-            // Workaround for JDK-8313816
-            VMMethod* vm_method = VMMethod::fromMethodID(method);
-            if (vm_method == NULL || vm_method->id() == NULL) {
-                return false;
-            }
+        if (VMMethod::isStaleMethodId(method)) {
+            return false;
         }
 
-        jvmtiEnv* jvmti = VM::jvmti();
-
         jclass method_class = NULL;
         char* class_name = NULL;
         char* method_name = NULL;
         char* method_sig = NULL;
 
-        if (jvmti->GetMethodName(method, &method_name, &method_sig, NULL) == 0 &&
-            jvmti->GetMethodDeclaringClass(method, &method_class) == 0 &&
-            jvmti->GetClassSignature(method_class, &class_name, NULL) == 0) {
+        jvmtiEnv* jvmti = VM::jvmti();
+        jvmtiError err;
+
+        if ((err = jvmti->GetMethodName(method, &method_name, &method_sig, NULL)) == 0 &&
+            (err = jvmti->GetMethodDeclaringClass(method, &method_class)) == 0 &&
+            (err = jvmti->GetClassSignature(method_class, &class_name, NULL)) == 0) {
             mi->_class = _classes->lookup(class_name + 1, strlen(class_name) - 2);
             mi->_name = _symbols.lookup(method_name);
             mi->_sig = _symbols.lookup(method_sig);
-        } else {
-            mi->_class = _classes->lookup("");
-            mi->_name = _symbols.lookup("jvmtiError");
-            mi->_sig = _symbols.lookup("()L;");
         }
 
         if (method_class) {
@@ -206,6 +201,10 @@ class Lookup {
         jvmti->Deallocate((unsigned char*)method_name);
         jvmti->Deallocate((unsigned char*)class_name);
 
+        if (err != 0) {
+            return false;
+        }
+
         if (first_time && jvmti->GetMethodModifiers(method, &mi->_modifiers) != 0) {
             mi->_modifiers = 0;
         }
@@ -260,6 +259,10 @@ class Lookup {
                 fillNativeMethodInfo(mi, buf, NULL);
             } else if (frame.bci == BCI_ERROR) {
                 fillNativeMethodInfo(mi, (const char*)method, NULL);
+            } else if (frame.bci == BCI_CPU) {
+                char buf[32];
+                snprintf(buf, sizeof(buf), "CPU-%d", ((int)(uintptr_t)method) & 0x7fff);
+                fillNativeMethodInfo(mi, buf, NULL);
             } else {
                 fillJavaClassInfo(mi, (uintptr_t)method);
             }
@@ -391,6 +394,12 @@ class Buffer {
         put(v, len);
     }
 
+    void putByteString(const char* v, u32 len) {
+        put8(5); // STRING_ENCODING_LATIN1_BYTE_ARRAY
+        putVar32(len);
+        put(v, len);
+    }
+
     void put8(int offset, char v) {
         _data[offset] = v;
     }
@@ -767,11 +776,11 @@ class Recording {
         buf->putVar32(0);
         buf->putVar32(0x7fffffff);  // must not clash with JFR metadata ID, or 'jfr print' will break
 
-        std::vector<std::string>& strings = JfrMetadata::strings();
+        const Index& strings = JfrMetadata::strings();
         buf->putVar32(strings.size());
-        for (int i = 0; i < strings.size(); i++) {
-            buf->putUtf8(strings[i].c_str());
-        }
+        strings.forEachOrdered([&] (const std::string& s) {
+            buf->putUtf8(s.c_str());
+        });
 
         writeElement(buf, JfrMetadata::root());
 
@@ -808,6 +817,7 @@ class Recording {
     }
 
     void writeSettings(Buffer* buf, Arguments& args) {
+        assert(args._cstack < sizeof(SETTING_CSTACK) / sizeof(char*));
         writeStringSetting(buf, T_ACTIVE_RECORDING, "version", PROFILER_VERSION);
         writeStringSetting(buf, T_ACTIVE_RECORDING, "engine", Profiler::instance()->_engine->type());
         writeStringSetting(buf, T_ACTIVE_RECORDING, "cstack", SETTING_CSTACK[args._cstack]);
@@ -993,7 +1003,7 @@ class Recording {
         buf->putVar32(0);
         buf->putVar32(1);
 
-        buf->putVar32(10);
+        buf->putVar32(11);
 
         Lookup lookup(&_method_map, Profiler::instance()->classMap());
         writeFrameTypes(buf);
@@ -1005,6 +1015,10 @@ class Recording {
         writeClasses(buf, &lookup);
         writePackages(buf, &lookup);
         writeSymbols(buf, &lookup);
+        writeUserEventTypes(buf);
+        // Write log levels last. The order does not affect the JFR's validity,
+        // but log levels have an easily-visible format that makes it easy
+        // to see if a JFR file has been accidentally truncated.
         writeLogLevels(buf);
     }
 
@@ -1193,6 +1207,18 @@ class Recording {
         }
     }
 
+    void writeUserEventTypes(Buffer* buf) {
+        std::map<u32, const char*> events;
+        UserEvents::collect(events);
+
+        writePoolHeader(buf, T_USER_EVENT_TYPE, events.size());
+        for (std::map<u32, const char*>::const_iterator it = events.begin(); it != events.end(); ++it) {
+            flushIfNeeded(buf, RECORDING_BUFFER_LIMIT - MAX_STRING_LENGTH);
+            buf->putVar32(it->first);
+            buf->putUtf8(it->second);
+        }
+    }
+
     void recordExecutionSample(Buffer* buf, int tid, u32 call_trace_id, ExecutionEvent* event) {
         int start = buf->skip(1);
         buf->put8(T_EXECUTION_SAMPLE);
@@ -1250,6 +1276,24 @@ class Recording {
         buf->put8(start, buf->offset() - start);
     }
 
+    void recordUserEvent(Buffer* buf, int tid, UserEvent* event) {
+        // estimate of size of non-string fields of this event
+        const size_t event_non_string_size_limit = 64;
+        // When calling recordUserEvent, the buffer can be up to RECORDING_BUFFER_LIMIT bytes full.
+        // Check that the buffer is not exceeded.
+        static_assert(RECORDING_BUFFER_LIMIT + event_non_string_size_limit + ASPROF_MAX_JFR_EVENT_LENGTH
+            <= RECORDING_BUFFER_SIZE, "output must fit within recording buffer");
+
+        int start = buf->skip(5);
+        buf->put8(T_USER_EVENT);
+        buf->putVar64(event->_start_time);
+        buf->putVar32(tid);
+        buf->putVar32(event->_type);
+        buf->putByteString((const char*)event->_data,
+            event->_len > ASPROF_MAX_JFR_EVENT_LENGTH ? ASPROF_MAX_JFR_EVENT_LENGTH : event->_len);
+        buf->putVar32(start, buf->offset() - start);
+    }
+
     void recordLiveObject(Buffer* buf, int tid, u32 call_trace_id, LiveObject* event) {
         int start = buf->skip(1);
         buf->put8(T_LIVE_OBJECT);
@@ -1528,6 +1572,9 @@ void FlightRecorder::recordEvent(int lock_index, int tid, u32 call_trace_id,
             case PROFILING_WINDOW:
                 _rec->recordWindow(buf, tid, (ProfilingWindow*)event);
                 break;
+            case USER_EVENT:
+                _rec->recordUserEvent(buf, tid, (UserEvent*)event);
+                break;
         }
         _rec->flushIfNeeded(buf);
         _rec->addThread(tid);
