@@ -9,6 +9,7 @@
 #include <jvmti.h>
 #include <stdint.h>
 #include <string.h>
+#include <type_traits>
 #include "codeCache.h"
 
 
@@ -25,6 +26,7 @@ class VMStructs {
     static bool _has_class_loader_data;
     static bool _has_native_thread_id;
     static bool _has_perm_gen;
+    static bool _can_dereference_jmethod_id;
     static bool _compact_object_headers;
 
     static int _klass_name_offset;
@@ -53,6 +55,8 @@ class VMStructs {
     static int _frame_complete_offset;
     static int _code_offset;
     static int _data_offset;
+    static int _mutable_data_offset;
+    static int _relocation_size_offset;
     static int _scopes_pcs_offset;
     static int _scopes_data_offset;
     static int _nmethod_name_offset;
@@ -137,6 +141,7 @@ class VMStructs {
 
     template<typename T>
     static T align(const void* ptr) {
+        static_assert(std::is_pointer<T>::value, "T must be a pointer type");
         return (T)((uintptr_t)ptr & ~(sizeof(T) - 1));
     }
 
@@ -345,10 +350,7 @@ class VMThread : VMStructs {
 
     static int nativeThreadId(JNIEnv* jni, jthread thread);
 
-    int osThreadId() {
-        const char* osthread = *(const char**) at(_thread_osthread_offset);
-        return osthread != NULL ? *(int*)(osthread + _osthread_id_offset) : -1;
-    }
+    int osThreadId();
 
     int state() {
         return _thread_state_offset >= 0 ? *(int*) at(_thread_state_offset) : 0;
@@ -384,12 +386,18 @@ class VMThread : VMStructs {
 
 class VMMethod : VMStructs {
   public:
-    static VMMethod* fromMethodID(jmethodID id) {
-        return *(VMMethod**)id;
-    }
-
     jmethodID id();
 
+    // Performs extra validation when VMMethod comes from incomplete frame
+    jmethodID validatedId();
+
+    // Workaround for JDK-8313816
+    static bool isStaleMethodId(jmethodID id) {
+        if (!_can_dereference_jmethod_id) return false;
+        VMMethod* vm_method = *(VMMethod**)id;
+        return vm_method == NULL || vm_method->id() == NULL;
+    }
+
     const char* bytecode() {
         return *(const char**) at(_method_constmethod_offset) + _constmethod_size;
     }
@@ -488,7 +496,11 @@ class NMethod : VMStructs {
     }
 
     VMMethod** metadata() {
-        if (_data_offset > 0) {
+        if (_mutable_data_offset >= 0) {
+            // Since JDK 25
+            return (VMMethod**) (*(char**) at(_mutable_data_offset) + *(int*) at(_relocation_size_offset));
+        } else if (_data_offset > 0) {
+            // since JDK 23
             return (VMMethod**) at(*(int*) at(_data_offset) + *(unsigned short*) at(_nmethod_metadata_offset));
         }
         return (VMMethod**) at(*(int*) at(_nmethod_metadata_offset));
